import { __extends, __assign } from 'tslib';
import React__default from 'react';
import { LayoutGroupContext } from '../../../context/LayoutGroupContext.js';
import { correctBorderRadius } from '../../../projection/styles/scale-border-radius.js';
import { correctBoxShadow } from '../../../projection/styles/scale-box-shadow.js';
import { addScaleCorrector } from '../../../projection/styles/scale-correction.js';
import { HTMLProjectionNode } from '../../../projection/node/HTMLProjectionNode.js';

var MeasureLayout = /** @class */ (function (_super) {
    __extends(MeasureLayout, _super);
    function MeasureLayout() {
        return _super !== null && _super.apply(this, arguments) || this;
    }
    /**
     * This only mounts projection nodes for components that
     * need measuring, we might want to do it for all components
     * in order to incorporate transforms
     */
    MeasureLayout.prototype.componentDidMount = function () {
        var visualElement = this.props.visualElement;
        var projection = visualElement.projection;
        addScaleCorrector(defaultScaleCorrectors);
        projection.setOptions({
            onProjectionUpdate: function () { return visualElement.scheduleRender(); },
        });
        var group = this.context.group;
        if (group)
            group.add(projection);
    };
    MeasureLayout.prototype.getSnapshotBeforeUpdate = function (prevProps) {
        var _a;
        var _b = this.props, layoutDependency = _b.layoutDependency, visualElement = _b.visualElement;
        if (prevProps.layoutDependency !== layoutDependency ||
            layoutDependency === undefined) {
            (_a = visualElement.projection) === null || _a === void 0 ? void 0 : _a.willUpdate();
        }
        return null;
    };
    MeasureLayout.prototype.componentDidUpdate = function () {
        var _a;
        (_a = this.props.visualElement.projection) === null || _a === void 0 ? void 0 : _a.root.didUpdate();
    };
    MeasureLayout.prototype.componentWillUnmount = function () {
        var group = this.context.group;
        if (group)
            group.remove(this.props.visualElement.projection);
    };
    MeasureLayout.prototype.render = function () {
        var _a;
        var visualElement = this.props.visualElement;
        if (!visualElement.projection) {
            var parent_1 = (_a = visualElement.parent) === null || _a === void 0 ? void 0 : _a.projection;
            visualElement.projection = new HTMLProjectionNode(parent_1);
        }
        return null;
    };
    MeasureLayout.contextType = LayoutGroupContext;
    return MeasureLayout;
}(React__default.Component));
var defaultScaleCorrectors = {
    borderRadius: __assign(__assign({}, correctBorderRadius), { applyTo: [
            "borderTopLeftRadius",
            "borderTopRightRadius",
            "borderBottomLeftRadius",
            "borderBottomRightRadius",
        ] }),
    borderTopLeftRadius: correctBorderRadius,
    borderTopRightRadius: correctBorderRadius,
    borderBottomLeftRadius: correctBorderRadius,
    borderBottomRightRadius: correctBorderRadius,
    boxShadow: correctBoxShadow,
};

export { MeasureLayout };
