import { ResolvedValues } from "../../render/types";
import { SubscriptionManager } from "../../utils/subscription-manager";
import { Box, Delta, Point } from "../geometry/types";
import { IProjectionNode, LayoutUpdateData, LayoutUpdateHandler, ProjectionNodeConfig, ProjectionNodeOptions, Snapshot } from "./types";
export declare function createProjectionNode<I>({ attachResizeListener, defaultParent, measureScroll, measureViewportBox, resetTransform, }: ProjectionNodeConfig<I>): {
    new (parent?: IProjectionNode<unknown> | undefined): {
        instance: I;
        root: IProjectionNode;
        parent: IProjectionNode;
        path: IProjectionNode[];
        children: Set<IProjectionNode<unknown>>;
        options: ProjectionNodeOptions;
        snapshot: Snapshot | undefined;
        layout: Box | undefined;
        layoutCorrected: Box;
        scroll?: Point | undefined;
        isLayoutDirty: boolean;
        shouldResetTransform: boolean;
        treeScale: Point;
        targetDelta?: Delta | undefined;
        projectionDelta?: Delta | undefined;
        target?: Box | undefined;
        layoutWillUpdateListeners?: SubscriptionManager<VoidFunction> | undefined;
        layoutDidUpdateListeners?: SubscriptionManager<LayoutUpdateHandler> | undefined;
        destructor(): void;
        /**
         * Lifecycles
         */
        mount(instance: I): void;
        willUpdate(shouldNotifyListeners?: boolean): void;
        didUpdate(): void;
        scheduleUpdateProjection(): void;
        updateProjection: () => void;
        /**
         * Update measurements
         */
        updateSnapshot(): void;
        updateLayout(): void;
        updateScroll(): void;
        resetTransform(): void;
        measure(): Box;
        removeElementScroll(box: Box): Box;
        /**
         *
         */
        setTargetDelta(delta: Delta): void;
        setOptions(options: ProjectionNodeOptions): void;
        /**
         * Frame calculations
         */
        resolveTargetDelta(): void;
        calcProjection(): void;
        getProjectionStyles(latest: ResolvedValues): ResolvedValues;
        /**
         * Events
         *
         * TODO Replace this with a key-based lookup
         */
        onLayoutWillUpdate(handler: VoidFunction): () => void;
        onLayoutDidUpdate(handler: (data: LayoutUpdateData) => void): () => void;
    };
};
